/*
 * Copyright (c) 2007 by Brian Marick.
 * Inspired by the Java version of the Hierarchical Data format, from the
 * University of Illinois NCSA and The HDF Group (http://www.hdfgroup.org/).
 * This code is intended for use only for training in test-driven design.
 * See LICENSE.txt
 */

package exampler.hdf.internal.converters;

import exampler.hdf.CompoundDataset;
import exampler.hdf.internal.Like;
import exampler.hdf.internal.NumberDescription;

import java.lang.reflect.Constructor;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.HashMap;

// TODO: The code within Member.holding could easily be generated from
// a table that maps FLOAT_TYPE,4 to {asFloatBuffer, float}

abstract public class Converter {
    private final NumberDescription description;

    abstract public boolean equalData(Object one, Object other);
    abstract public Object convertToJavaFormat(ByteBuffer rawBytes, long lengthOfResult);
    abstract public void overwriteAsBytes(Object javaish, ByteBuffer buf);

    void setByteOrder(ByteBuffer buf) {
        buf.order(description.getByteOrder());
    }


    static private final HashMap classMap = new HashMap();
    static {
        map(Like.integer(3),  IntArray_3Byte_Converter.class);
        map(Like.integer(4),  IntArray_4Byte_Converter.class);
        map(Like.integer(8),  LongArray_8Byte_Converter.class);
        map(Like.floating(4), FloatArray_4Byte_Converter.class);
        map(Like.floating(8), DoubleArray_8Byte_Converter.class);
    }


    public static Converter of(NumberDescription description) {
        try {
            Class klass = (Class) classMap.get(description.withoutByteOrder());
            Constructor c = klass.getConstructor(NumberDescription.class);
            return (Converter) c.newInstance(description);
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    public static Converter of(Object o) {
        return of(o, CompoundDataset.DEFAULT_BYTE_ORDER);
    }

    public static Converter of(Object o, ByteOrder byteOrder) {
        assert o.getClass().isArray();

        if (isA(o, "int"))      return Converter.of(new int[0], byteOrder);
        if (isA(o, "long"))     return Converter.of(new long[0], byteOrder);
        if (isA(o, "float"))    return Converter.of(new float[0], byteOrder);
        if (isA(o, "double"))   return Converter.of(new double[0], byteOrder);

        assert false;
        return null;
    }

    public static Converter of(int[] ints) { return of(ints, CompoundDataset.DEFAULT_BYTE_ORDER); }
    @SuppressWarnings({"UnusedParameters"})
    public static Converter of(int[] ints, ByteOrder byteOrder) {
        return of(Like.integer(inBytes(Integer.SIZE), byteOrder));
    }

    public static Converter of(long[] longs) { return of(longs, CompoundDataset.DEFAULT_BYTE_ORDER); }
    @SuppressWarnings({"UnusedParameters"})
    public static Converter of(long[] longs, ByteOrder byteOrder) {
        return of(Like.integer(inBytes(Long.SIZE), byteOrder));
    }

    public static Converter of(float[] floats) { return of(floats, CompoundDataset.DEFAULT_BYTE_ORDER); }
    @SuppressWarnings({"UnusedParameters"})
    public static Converter of(float[] floats, ByteOrder byteOrder) {
        return of(Like.floating(inBytes(Float.SIZE), byteOrder));
    }

    public static Converter of(double[] doubles) { return of(doubles, CompoundDataset.DEFAULT_BYTE_ORDER); }
    @SuppressWarnings({"UnusedParameters"})
    public static Converter of(double[] doubles, ByteOrder byteOrder) {
        return of(Like.floating(inBytes(Double.SIZE), byteOrder));
    }


    public Converter(NumberDescription description) {
        this.description = description;
    }


    public int getBytesPerElement() { return description.getSizeInBytes(); }
    public String getElementType() { return description.getType(); }
    public ByteOrder getByteOrder() { return description.getByteOrder(); }

    private static int inBytes(int bits) {
        return bits / Byte.SIZE;
    }


    private static boolean isA(Object o, String elementType) {
        return o.getClass().getComponentType().getName().equals(elementType);
    }


    private static void map(NumberDescription key, Class klass) {
        classMap.put(key.withoutByteOrder(), klass);
    }


    public NumberDescription getDescription() {
        return description;
    }

}
