/*
 * Copyright (c) 2007 by Brian Marick.
 * Inspired by the Java version of the Hierarchical Data format, from the
 * University of Illinois NCSA and The HDF Group (http://www.hdfgroup.org/).
 * This code is intended for use only for training in test-driven design.
 * See LICENSE.txt
 */

package exampler.hdf.internal.diskAccess;

import exampler.hdf.CompoundDataset;
import exampler.hdf.internal.CompoundDatasetFAQ;
import exampler.hdf.internal.Errors;
import exampler.hdf.internal.NumberDescription;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.nio.ByteOrder;

/**
 * All knowledge of disk layout lives here.
 */

public class PileLayout implements CompoundDatasetFAQ {

    private long dimensions = -1;
    private int numberOfMembers = -1;
    private int[] elementTypeIds = null;
    private int[] elementSizes = null;
    private int[] elementByteOrders = null;

    private PileLayout(long dimensions, int numberOfMembers) {
        this.dimensions = dimensions;
        this.numberOfMembers = numberOfMembers;
        elementTypeIds = new int[numberOfMembers];
        elementSizes = new int[numberOfMembers];
        elementByteOrders = new int[numberOfMembers];
    }

    public static PileLayout createFrom(CompoundDatasetFAQ oracle) {
        PileLayout retval = new PileLayout(oracle.tellMemberDimensions(),
                                           oracle.tellHowManyMembers());
        for (int i = 0; i < oracle.tellHowManyMembers(); i++) {
            retval.elementTypeIds[i] = diskType(oracle.tellTypeOf(i));
            retval.elementSizes[i] = oracle.tellSizeInBytesOf(i);
            retval.elementByteOrders[i] = h5ByteOrder(oracle.tellByteOrderOf(i));
        }
        return retval;
    }

    public static PileLayout createFrom(DataInput stream) {
        try {
            long dimensions = stream.readLong();
            int numberOfMembers = stream.readInt();
            PileLayout retval = new PileLayout(dimensions, numberOfMembers);
            retval.dimensions = dimensions;
            retval.numberOfMembers = numberOfMembers;
            retval.elementTypeIds = readInts(stream, numberOfMembers);
            retval.elementSizes = readInts(stream, numberOfMembers);
            retval.elementByteOrders = readInts(stream, numberOfMembers);
            return retval;
        } catch (IOException e) {
            Errors.panic(e.getMessage());
            return null;
        }
    }

    public void write(DataOutput stream) {
        try {
            stream.writeLong(dimensions);
            stream.writeInt(numberOfMembers);
            writeInts(stream, elementTypeIds);
            writeInts(stream, elementSizes);
            writeInts(stream, elementByteOrders);
        } catch (IOException e) {
            Errors.panic(e.getMessage());
        }
    }



    // Accessors - this is a FAQ

    public long tellMemberDimensions() { return dimensions; }
    public int tellHowManyMembers() { return numberOfMembers; }

    public String tellTypeOf(int member) { return type(elementTypeIds[member]); }
    public int tellSizeInBytesOf(int i) { return elementSizes[i]; }
    public ByteOrder tellByteOrderOf(int i) { return byteOrder(elementByteOrders[i]); }
    public long tellHowManyBytesOfMetadata() {
        return memberDataOffset();
    }

    public long tellHowManyBytesDataSetRequires() {
        long retval = memberDataOffset();
        for (int size : elementSizes) {
            retval += dimensions * size;
        }
        return retval;
    }

    public NumberDescription describeMember(int member) {
        return new NumberDescription(tellTypeOf(member),
                                     tellSizeInBytesOf(member),
                                     tellByteOrderOf(member));

    }



    // Disk layout details
    private static final int DISK_INTEGER_TYPE = 0x01111111;
    private static final int DISK_FLOAT_TYPE =   0x0FFFFFFF;

    private static final int DISK_BIG_ENDIAN = 0xB16B16;
    private static final int DISK_LITTLE_ENDIAN = 0x109109;



    private static final int DIMENSION_OFFSET = 0;
    static final int DIMENSION_SIZE = 8;

    @SuppressWarnings({"PointlessArithmeticExpression"})
    private static final int NUMBER_OF_MEMBERS_OFFSET = DIMENSION_OFFSET + DIMENSION_SIZE;
    static final int NUMBER_OF_MEMBERS_SIZE = 4;

    private static final int MEMBER_TYPE_OFFSET = NUMBER_OF_MEMBERS_OFFSET + NUMBER_OF_MEMBERS_SIZE;
    static final int MEMBER_TYPE_ELEMENT_SIZE = 4;

    private int memberSizeOffset(int numberOfMembers) {
        return MEMBER_TYPE_OFFSET + MEMBER_TYPE_ELEMENT_SIZE * numberOfMembers;
    }
    static final int MEMBER_SIZE_ELEMENT_SIZE = 4;

    private int memberByteOrderOffset(int numberOfMembers) {
        return memberSizeOffset(numberOfMembers) + MEMBER_SIZE_ELEMENT_SIZE * numberOfMembers;
    }
    static final int MEMBER_BYTE_ORDER_ELEMENT_SIZE = 4;

    int memberDataOffset() {
        return memberByteOrderOffset(numberOfMembers) + MEMBER_BYTE_ORDER_ELEMENT_SIZE * numberOfMembers;
    }



    // Low-level conversions
    private static int h5ByteOrder(ByteOrder byteOrder) {
        return byteOrder.equals(ByteOrder.BIG_ENDIAN) ? DISK_BIG_ENDIAN : DISK_LITTLE_ENDIAN;
    }

    private static ByteOrder byteOrder(int h5ByteOrder) {
        return h5ByteOrder == DISK_BIG_ENDIAN ? ByteOrder.BIG_ENDIAN : ByteOrder.LITTLE_ENDIAN;
    }


    private static int diskType(String type) {
        if (type.equals(CompoundDataset.INTEGER_TYPE)) {
            return DISK_INTEGER_TYPE;
        } else if (type.equals(CompoundDataset.FLOAT_TYPE)) {
            return DISK_FLOAT_TYPE;
        }
        Errors.panic("Unknown type %s", type);
        return -1;
    }

    private static String type(int diskType) {
        switch(diskType) {
            case DISK_INTEGER_TYPE: return CompoundDataset.INTEGER_TYPE;
            case DISK_FLOAT_TYPE: return CompoundDataset.FLOAT_TYPE;
            default: Errors.panic("Odd type %d", diskType);
        }
        return null;
    }



    // Util

    private static int[] readInts(DataInput stream, int count) throws IOException {
        int[] ints = new int[count];
        for (int i = 0; i < ints.length; i++) {
            ints[i] = stream.readInt();
        }
        return ints;
    }

    private static void writeInts(DataOutput stream, int[] ints) throws IOException {
        for (int i : ints) {
            stream.writeInt(i);
        }

    }


}
