/*
 * Copyright (c) 2007 by Brian Marick.
 * Inspired by the Java version of the Hierarchical Data format, from the
 * University of Illinois NCSA and The HDF Group (http://www.hdfgroup.org/).
 * This code is intended for use only for training in test-driven design.
 * See LICENSE.txt
 */


package exampler.hdf.internal.diskAccess;

import exampler.hdf.internal.CompoundDatasetFAQ;
import exampler.hdf.internal.Errors;
import exampler.hdf.internal.NumberDescription;
import exampler.hdf.internal.converters.Converter;

import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel;

public class PileOfBytes {
    private final RandomAccessFile stream;
    private MappedByteBuffer fileContents;
    private final PileLayout layout;

    static public PileOfBytes openFileForReading(String filename) {
        PileOfBytes reader = null;
        try {
            RandomAccessFile stream = new RandomAccessFile(filename, "r");
            PileLayout layout = PileLayout.createFrom(stream);
            reader = new PileOfBytes(stream, layout);
            reader.map(FileChannel.MapMode.READ_ONLY,
                       layout.tellHowManyBytesDataSetRequires());
        } catch (Exception e) {
            Errors.panic(e);
        }
        return reader;
    }

    static public PileOfBytes openFileForWriting(String filename, CompoundDatasetFAQ oracle) {
        PileLayout layout = PileLayout.createFrom(oracle);
        PileOfBytes writer = null;
        try {
            RandomAccessFile stream = new RandomAccessFile(filename, "rwd");
            layout.write(stream);
            writer = new PileOfBytes(stream, layout);
            writer.map(FileChannel.MapMode.READ_WRITE, layout.tellHowManyBytesDataSetRequires());
        } catch (Exception e) {
            Errors.panic(e);
        }
        return writer;
    }

    public void close() {
        try {
            stream.close();
        } catch (IOException e) {
            Errors.panic(e);
        }
    }


    private PileOfBytes(RandomAccessFile stream, PileLayout layout) {
        this.stream = stream;
        this.layout = layout;
    }



    public void overwriteAsBytes(Object memberData, int i, Converter converter) {
        converter.overwriteAsBytes(memberData, getBufferForMemberRawData(i));
    }

    public Object createDataFromBytes(int i, Converter converter) {
        ByteBuffer buf = getBufferForMemberRawData(i);
        return converter.convertToJavaFormat(buf, layout.tellMemberDimensions());
    }


    // Give some access to layout.
    public long tellMemberDimensions() {
        return layout.tellMemberDimensions();
    }

    public int tellHowManyMembers() {
        return layout.tellHowManyMembers();
    }

    public NumberDescription describeMember(int i) {
        return layout.describeMember(i);
    }
    


    // Util

    private void map(FileChannel.MapMode mapMode, long size) {
        try {
            FileChannel c = stream.getChannel();
            fileContents = c.map(mapMode, 0, size);
        } catch (Exception e) {
            Errors.panic(e);
        }
    }

    private ByteBuffer getBufferForMemberRawData(int index) {
        setCursorForMemberRawData(index);
        ByteBuffer retval = fileContents.slice();
        retval.limit((int)bytesForMember(index)); // TODO: Java has small files (int-sized)
        return retval;
    }

    private void setCursorForMemberRawData(int index) {
        int cursor = layout.memberDataOffset();
        for (int i = 0; i < index; i++) {
            cursor += bytesForMember(i);
        }
        fileContents.position(cursor);
    }

    private long bytesForMember(int i) {
        return layout.tellMemberDimensions() * layout.tellSizeInBytesOf(i);
    }

}
