/*
 * Copyright (c) 2007 by Brian Marick.
 * Inspired by the Java version of the Hierarchical Data format, from the
 * University of Illinois NCSA and The HDF Group (http://www.hdfgroup.org/).
 * This code is intended for use only for training in test-driven design.
 * See LICENSE.txt
 */

package exampler.hdf.internal.converters;

import junit.framework.TestCase;
import exampler.hdf.internal.Like;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.Arrays;

public class FloatArray_4Byte_ConverterTest extends TestCase {
    public void test_can_compare_arrays_of_floats() {
        Converter converter = Converter.of(new float[3]);
        float[] one = {1.0f, 2.0f, 3.0f };
        float[] same = {1.0f, 2.0f, 3.0f };
        float[] diff = {10.0f, 2.0f, 3.0f };
        assertTrue(converter.equalData(one, same));
        assertFalse(converter.equalData(one, diff));
    }

    // to bytes
    private final float[] floats = new float[]{3.3f, 2.0f};
    private final byte[] bigEndianBytes =    { 64, 83, 51, 51, 64, 0, 0, 0 };
    private final byte[] littleEndianBytes = { 51, 51, 83, 64, 0, 0, 0, 64 };



    private void toBytes(Converter converter, float[] floats, byte[] expected) {
        byte[] destination = {  5, 5, 5, 5, 5, 5, 5, 5 };
        ByteBuffer buf = ByteBuffer.wrap(destination);
        converter.overwriteAsBytes(floats, buf);

        assertTrue(Arrays.equals(expected, destination));
    }

    public void test_conversion_of_java_floats_to_big_endian_bytes() {
        toBytes(Converter.of(floats, ByteOrder.BIG_ENDIAN), 
                floats,
                bigEndianBytes);
    }

    public void test_conversion_of_java_floats_to_little_endian_bytes() {
        toBytes(Converter.of(floats, ByteOrder.LITTLE_ENDIAN),
                floats,
                littleEndianBytes);
    }


    public void test_conversion_of_java_floats_is_big_endian_by_default() {
        toBytes(Converter.of(floats),
                floats,
                bigEndianBytes);
    }

    // From bytes

    private void fromBytes(Converter converter, byte[] bytes, float[] expected) {
        ByteBuffer buf = ByteBuffer.wrap(bytes);
        assertTrue(Arrays.equals(expected,
                                 (float[]) converter.convertToJavaFormat(buf, expected.length)));
    }

    public void test_conversion_of_a_big_endian_raw_buffer_to_floats() {
        fromBytes(Converter.of(Like.floating(4, ByteOrder.BIG_ENDIAN)),
                  bigEndianBytes,
                  floats);
    }


    public void test_conversion_of_a_little_endian_raw_buffer_to_floats() {
        fromBytes(Converter.of(Like.floating(4, ByteOrder.LITTLE_ENDIAN)),
                  littleEndianBytes,
                  floats);
    }

    public void test_conversion_of_a_raw_buffer_uses_big_endian_by_default() {
        fromBytes(Converter.of(Like.floating(4)),
                  bigEndianBytes,
                  floats);
    }




}
